/* symtab.c   R1 compiler
	P. Chongstitvatana
	19 Aug 97
*/


#include "compile.h"

#define		PRIVATE		static

sym_entry symtab[TABLESIZE];
int lv = 0;

PRIVATE int next_location = 1;
PRIVATE int hashtable[TABLESIZE];
PRIVATE int localList[LOCALSIZE];

/* return 0 not found, index to symtab if found
   h is index to hashtable */
int lookup( char *Name, int *h )
{
  int start, try1, hash_try;
  try1 = hash(Name);
  start = try1;
  while( hash_try = hashtable[try1] ) {  /* found something */
    if( !strcmp( symtab[hash_try].name, Name) ) {
      *h = try1;
      return hash_try;
    }
    if( ++try1 >= TABLESIZE ) try1 = 0; /* wrap */
    if( try1 == start ) seterror("symbol table full");
  }
  *h = try1;
  return 0;
}

/* for formal params, mask global */
insertLocal( char *name )
{
  int a,h;
  a = lookup(name, &h);
  enterLocal(name, h);
  if( a ) {
    if( symtab[a].type == tyLOCAL )
      seterror("duplicate identifiers in formal parameter list");
    symtab[hashtable[h]].arg = a;
  }
}

int enterGlobal( char *name, int ty, int ref, int arg, int h )
{
  hashtable[h] = newSym(name,ty,ref,arg);
  return hashtable[h];
}

/* check an identifier for type and ref return index to symtab */
int chkLocal( char *name, int *type)
{
  int a,h,ref;
  a = lookup(name,&h);
  if( a == 0 ) {                   /* not found in symtab */
    *type = tyLOCAL;
    enterLocal(name,h);
    ref = hashtable[h];
  } else {
    *type = symtab[a].type;
    ref = a;
  }
  return ref;
}

cleanLocal()
{
  int i,h,a;
  if( lv == 0 ) return;
  next_location = hashtable[localList[0]];
  for( i=0; i<lv; i++ ) {
    h = localList[i];
    a = hashtable[h];
    hashtable[h] = symtab[a].arg;
  }
  lv = 0;
}

putSymtabArg( int i, int arg )
{
  symtab[i].arg = arg;
}

#ifdef DEBUG
dumpSymTab()
{
  int i;
  sym_entry *s;
  printf("symbol table %d entries\n%s\n",next_location-1,
	 "      name type ref arg");
  for( i=1; i<next_location; i++ ) {
    s = &symtab[i];
    printf("  %d %-8s %3d %3d %3d\n",
      i,s->name,s->type,s->ref,s->arg);
  }
}
#endif

/* from Aho dragon book, due to P.J. Weinberger'c C compiler */
PRIVATE hash( char *s )
{
  char *ss = s;
  unsigned long h,g;
  h = 0;
  for( ; *ss != '\0'; ss++ ) {
    h = (h << 4) + *ss;
    if ( g = h & 0xf0000000) {
      h ^= g >> 24;
      h ^= g;
    }
  }
  return (int) ( h % TABLESIZE );
}

/* return index to the new entry in symtab */
PRIVATE int newSym( char *name, int type, int ref, int arg)
{
  sym_entry  *s;
  s = &symtab[next_location];
  next_location++;
  strcpy(s->name,name);
  s->type = type;
  s->ref = ref;
  s->arg = arg;
  return next_location - 1;
}

/* insert a new local into symtab at hash index h
   keep track of local list for clean up */
PRIVATE enterLocal( char *name, int h )
{
  hashtable[h] = newSym(name,tyLOCAL,lv,0);
  localList[lv] = h;
  lv++;
}


